# -*- coding: utf-8 -*-

# Standard Library
import datetime
from dateutil.relativedelta import relativedelta

# Vendor Library
from sqlalchemy import and_, or_
from sqlalchemy.orm import load_only
from sqlalchemy.sql.expression import false, true

# Project Library
from kinder.core.declaration_status.models import DSS
from kinder.core.dict.models import GroupTypeEnumerate
from gisdo.alchemy_models import (
    AlchemyGroupType, AlchemyGroupWorkType,
    AlchemyDeclarationStatus, AlchemyDeclaration, AlchemyChildren)
from gisdo.algorithm import BaseEnrolledAlgorithm
from gisdo.constants import ALL
from gisdo.utils import AgeDeltas

from gisdo.algorithm.constants import (
    get_all_day_work_types, get_short_day_types, not_health_need)


class EnrolledCountByAge(BaseEnrolledAlgorithm):
    """Кол-во зачислений в Ф группы, без временно отчисленных детей"""

    def get_result_set(self, **kwargs):

        not_temporary_deduct_pupils = super(
            EnrolledCountByAge, self).get_result_set()
        groups_this_unit = self._get_groups()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id))
            .join(
                kids,
                kids.c.children_id ==
                not_temporary_deduct_pupils.c.children_id)
            .group_by(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .subquery()
        )


class EnrolledInvalidWithoutHealthNeed(BaseEnrolledAlgorithm):
    u"""Дети-инвалиды без ограниченных возможностей здоровья (30 <= 19.1)."""

    def get_result_set(self, **kwargs):
        not_temporary_deduct_pupils = super(
            EnrolledInvalidWithoutHealthNeed, self).get_result_set()
        groups_this_unit = self._get_groups()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id))
            .join(
                kids,
                and_(
                    kids.c.children_id ==
                    not_temporary_deduct_pupils.c.children_id,
                    kids.c.is_invalid == true()
                ))
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code != GroupTypeEnumerate.HEALTH,
                        AlchemyGroupType.code != GroupTypeEnumerate.COMP),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code != GroupTypeEnumerate.HEALTH,
                        AlchemyGroupType.code != GroupTypeEnumerate.COMP)
                ))
            .filter(kids.c.health_need_id is None)
            .subquery()
        )


class EnrolledInvalidWithHealthNeed(BaseEnrolledAlgorithm):
    u"""Дети-инвалиды c ограниченными возможностями здоровья (30.1 <= 19.1)."""

    def get_result_set(self, **kwargs):
        not_temporary_deduct_pupils = super(
            EnrolledInvalidWithHealthNeed, self).get_result_set()
        groups_this_unit = self._get_groups()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id))
            .join(
                kids,
                and_(
                    kids.c.children_id ==
                    not_temporary_deduct_pupils.c.children_id,
                    kids.c.is_invalid == true()
                ))
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id)
                ))
            .filter(
                or_(
                    or_(AlchemyGroupType.code == GroupTypeEnumerate.HEALTH,
                        AlchemyGroupType.code == GroupTypeEnumerate.COMP),
                    kids.c.health_need_id is not None
                ))
            .subquery()
        )


class EnrolledNoInvalidWithHealthNeed(BaseEnrolledAlgorithm):
    u"""Дети неинвалиды, имеющие потребность по здоровью.

    30.2 <= 19.1,
    30.1 + 30.2 >= 22.1 + 22.2 + 22.3 + 22.4 + 22.5 + 22.6 + 22.7 + 22.8
    """

    def get_result_set(self, **kwargs):
        not_temporary_deduct_pupils = super(
            EnrolledNoInvalidWithHealthNeed, self).get_result_set()
        groups_this_unit = self._get_groups()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id))
            .join(
                kids,
                and_(
                    kids.c.children_id == not_temporary_deduct_pupils.c.children_id,
                    kids.c.is_invalid == false()
                ))
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id)
                ))
            .filter(
                or_(
                    or_(AlchemyGroupType.code == GroupTypeEnumerate.HEALTH,
                        AlchemyGroupType.code == GroupTypeEnumerate.COMP),
                    kids.c.health_need_id is not None
                ))
            .subquery()
        )


class PredictableDecreaseContingent(EnrolledCountByAge):
    """Прогнозируемое уменьшение контингента

    - количество детей в текущих группах (реестр "Группы")
      возраст которых больше 6,5 лет
    - дата расчета возраста = текущая дата
    """

    def _kids(self, **kwargs):

        today = datetime.date.today()
        # Если ДР раньше этой даты, то ребенку > 6,5 лет.
        date_of_birth_border = today - relativedelta(years=6, months=6)

        kids = (
            self.session.query(
                AlchemyChildren.id.label("children_id"),
                AlchemyChildren.date_of_birth
            ).filter(
                AlchemyChildren.date_of_birth < date_of_birth_border)
            .subquery("kids"))

        return kids


class EnrolledCountByCare(BaseEnrolledAlgorithm):
    """Группы по присмотру и уходу"""

    def get_result_set(self, **kwargs):

        not_temporary_deduct_pupils = super(
            EnrolledCountByCare, self).get_result_set()
        groups_this_unit = self._get_groups()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_is_family == false(),
                        or_(
                            groups_this_unit.c.fact_group_is_care == true(),
                            groups_this_unit.c.fact_group_is_young_children ==
                            true())
                    ),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_is_family == false(),
                        or_(
                            groups_this_unit.c.plan_group_is_care == true(),
                            groups_this_unit.c.plan_group_is_young_children ==
                            true())
                    )
                )
            )
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.DEV),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.DEV)
                ))
            .join(
                kids,
                kids.c.children_id ==
                not_temporary_deduct_pupils.c.children_id)
            .group_by(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .subquery()
        )


class EnrolledCountByDisabled(BaseEnrolledAlgorithm):

    def get_result_set(self, health_need=None, **kwargs):

        groups_this_unit = self._get_groups()
        not_temporary_deduct_pupils = super(
            EnrolledCountByDisabled, self).get_result_set()
        kids = self._kids(**kwargs)

        _filter = ''
        if health_need:
            _filter = and_(
                or_(groups_this_unit.c.fact_group_health_need_id.in_(
                        health_need),
                    groups_this_unit.c.plan_group_health_need_id.in_(
                        health_need
                    )),
                or_(AlchemyGroupType.code == GroupTypeEnumerate.COMP,
                    and_(AlchemyGroupType.code == GroupTypeEnumerate.COMBI,
                         kids.c.health_need_id.isnot(None),
                         kids.c.health_need_id.notin_(not_health_need))),
            )

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(
                    groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id
                ))
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id, or_(
                            AlchemyGroupType.code == GroupTypeEnumerate.COMP,
                            AlchemyGroupType.code == GroupTypeEnumerate.COMBI)
                        ),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id, or_(
                            AlchemyGroupType.code == GroupTypeEnumerate.COMP,
                            AlchemyGroupType.code == GroupTypeEnumerate.COMBI)
                    ),
                )
            )
            .join(
                kids,
                kids.c.children_id ==
                not_temporary_deduct_pupils.c.children_id)
            .group_by(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth
            ).filter(
                _filter
            ).subquery()
        )


class EnrolledCountByFamily(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        groups_this_unit = self._get_groups()
        not_temporary_deduct_pupils = super(
            EnrolledCountByFamily, self).get_result_set()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(
                    groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id
                ))
            .join(
                kids,
                kids.c.children_id ==
                not_temporary_deduct_pupils.c.children_id)
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.DEV),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.DEV)
                ))
            .filter(
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_is_family == true()
                    ),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_is_family == true())
                ))
            .group_by(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .subquery())


class EnrolledCountByCorrect(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        groups_this_unit = self._get_groups()
        not_temporary_deduct_pupils = super(
            EnrolledCountByCorrect, self).get_result_set()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(
                    groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id
                ))
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.COMP
                    ),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.COMP
                    )
                )
            )
            .join(
                kids,
                kids.c.children_id ==
                not_temporary_deduct_pupils.c.children_id)
            .group_by(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .subquery())


class EnrolledCountByHealth(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        groups_this_unit = self._get_groups()
        not_temporary_deduct_pupils = super(
            EnrolledCountByHealth, self).get_result_set()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(
                    groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id
                ))
            .join(
                AlchemyGroupType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.HEALTH
                    ),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_type_id ==
                        AlchemyGroupType.id,
                        AlchemyGroupType.code == GroupTypeEnumerate.HEALTH
                    )
                )
            )
            .join(
                kids,
                kids.c.children_id ==
                not_temporary_deduct_pupils.c.children_id)
            .group_by(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .subquery()
        )


class EnrolledCountStay(BaseEnrolledAlgorithm):
    def get_result_set(self, group_work_types=None, **kwargs):
        groups_this_unit = self._get_groups()
        not_temporary_deduct_pupils = super(
            EnrolledCountStay, self).get_result_set()
        kids = self._kids(**kwargs)

        return (
            self.session.query(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .join(
                groups_this_unit,
                or_(
                    groups_this_unit.c.fact_group_id ==
                    not_temporary_deduct_pupils.c.grup_id,
                    groups_this_unit.c.plan_group_id ==
                    not_temporary_deduct_pupils.c.grup_id
                ))
            .join(
                AlchemyGroupWorkType,
                or_(
                    and_(
                        groups_this_unit.c.fact_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.fact_group_work_type_id ==
                        AlchemyGroupWorkType.id,
                        AlchemyGroupWorkType.code.in_(group_work_types)),
                    and_(
                        groups_this_unit.c.plan_group_id ==
                        not_temporary_deduct_pupils.c.grup_id,
                        groups_this_unit.c.plan_group_work_type_id ==
                        AlchemyGroupWorkType.id,
                        AlchemyGroupWorkType.code.in_(group_work_types))))
            .join(
                kids,
                kids.c.children_id ==
                not_temporary_deduct_pupils.c.children_id)
            .group_by(
                not_temporary_deduct_pupils.c.children_id,
                kids.c.date_of_birth)
            .subquery())


class EnrolledCountShortStay(EnrolledCountStay):
    def get_result_set(self, **kwargs):
        return super(EnrolledCountShortStay, self).get_result_set(
            get_short_day_types(), **kwargs)


class EnrolledCountFullDayStay(EnrolledCountStay):
    def get_result_set(self, **kwargs):
        return super(EnrolledCountFullDayStay, self).get_result_set(
            get_all_day_work_types(), **kwargs)


class EnrolledCountWantChangeDou(BaseEnrolledAlgorithm):

    def get_result_set(self, **kwargs):

        fact_group_this_unit = self._get_fact_group()
        not_temporary_deduct_pupils = super(
            EnrolledCountWantChangeDou, self).get_result_set()
        kids = self._kids(**kwargs)
        pupils = self.session.query(
            not_temporary_deduct_pupils).join(
                fact_group_this_unit).subquery()
        pupils = (
            self.session.query(
                kids.c.id.label(
                    'children_id'))
            .join(
                pupils, pupils.c.children_id == kids.c.id)
            .subquery())

        want_change_dou_status = self.session.query(
            AlchemyDeclarationStatus).filter(
                AlchemyDeclarationStatus.code ==
                DSS.WANT_CHANGE_DOU).subquery()
        declarations_want_change_dou = (
            self.session.query(
                AlchemyDeclaration)
            .join(want_change_dou_status)
            .join(pupils,
                  AlchemyDeclaration.children_id ==
                  pupils.c.children_id)
            .options(load_only('children_id'))
            .distinct()
            .subquery())

        return declarations_want_change_dou


class EnrolledCountWantChangeDouWithHN(EnrolledCountWantChangeDou):

    def _kids(self, **kwargs):

        age_range = kwargs.get('age_range', ALL)
        age_on_date = kwargs.get('on_date', AgeDeltas.calculate_date)
        down, up = AgeDeltas.get_category_deltas(age_range, age_on_date)
        kids = (
            self.session.query(
                AlchemyChildren.id.label("children_id"),
                AlchemyChildren.date_of_birth)
            .filter(
                AlchemyChildren.date_of_birth > down,
                AlchemyChildren.date_of_birth <= up,
                AlchemyChildren.health_need_id != None)
            .options(load_only('id'))
            .subquery("kids"))
        return kids
